// WorktypeID重複エラー
export class WorktypeIdAlreadyExistsError extends Error {
  constructor(message: string) {
    super(message);
    this.name = 'WorktypeIdAlreadyExistsError';
  }
}
// WorktypeID登録上限エラー
export class WorktypeIdMaxCountError extends Error {
  constructor(message: string) {
    super(message);
    this.name = 'WorktypeIdMaxCountError';
  }
}
// WorktypeID不在エラー
export class WorktypeIdNotFoundError extends Error {
  constructor(message: string) {
    super(message);
    this.name = 'WorktypeIdNotFoundError';
  }
}
// WorktypeID使用中エラー
export class WorktypeIdInUseError extends Error {
  constructor(message: string) {
    super(message);
    this.name = 'WorktypeIdInUseError';
  }
}
